/*
 * file CqFieldValue.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM 
 *
 * com.ibm.rational.wvcm.stp.cq.CqFieldValue
 *
 *  Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.rational.wvcm.stp.cq;

import static com.ibm.rational.wvcm.stpex.StpExBase.METADATA_NAMESPACE;

import java.util.EnumMap;

import javax.wvcm.WvcmException;

import com.ibm.rational.wvcm.stp.StpProperty;
import com.ibm.rational.wvcm.stp.cq.CqQuery.DisplayField;
import com.ibm.rational.wvcm.stp.cq.CqQuery.DisplayField.FieldType;
import com.ibm.rational.wvcm.stpex.StpExBase;
import com.ibm.rational.wvcm.stpex.StpExEnumeration;
import com.ibm.rational.wvcm.stpex.StpExEnumerationBase;


/**
 * The representation for the value of a record field as an extended StpProperty
 * object.
 */
public interface CqFieldValue<T>
    extends StpProperty<T>
{
    /**
     * An enumeration of the possible types of values that can be stored in a
     * record field. This enumeration is more precise than the StpProperty.Type
     * enumeration in specifying the type of a field value.
     */
    public enum ValueType implements StpExEnumeration
    {
        /** An initial value for ValueType variables */
        UNKNOWN(Type.UNKNOWN),

        /**
         * A single-line of text up to 254 characters. Leading and trailing
         * whitespace is trimmed when stored in a record.
         */
        SHORT_STRING(Type.STRING),

        /**
         * An arbitrarily long string of characters, possibly containing one or
         * more newline characters.
         */
        MULTILINE_STRING(Type.STRING),

        /** An integer in the range -9999999999 to 9999999999 */
        INTEGER(Type.INTEGER),

        /**
         * A GMT date/time between 12:00:01 AM January 1, 1970 and 11:59:59 PM
         * January 18, 2038.
         */
        DATE_TIME(Type.DATE_TIME),

        /** A proxy for a Resource of a specific ResourceValueType. */
        RESOURCE(Type.RESOURCE),

        /** A list of proxies for a Resource of a specific ResourceValueType */
        RESOURCE_LIST(Type.RESOURCE_LIST),

        /** A floating point value */
        FLOAT(Type.FLOAT),

        /** A field containing attachments */
        ATTACHMENT_LIST(Type.RESOURCE_LIST),

        /** A record id (alphanumeric) field */
        ID(Type.STRING),

        /** A field containing the name of the record's current state */
        STATE(Type.STRING),

        /** A field recording events in the history of the record */
        JOURNAL(Type.OBJECT),

        /** A database id (integer) field */
        DBID(Type.LONG),

        /** A field containing the type of the record's current state */
        STATE_TYPE(Type.STRING),

        /** A field containing the record's record type name */
        RECORD_TYPE(Type.STRING);

        /**
         * The StpProperty.Type that corresponds to this field type.
         * 
         * @return An StpProperty.Type enumerator specifying as closely as
         *         possible the Java representation of this type of value
         *         corresponding
         */
        public Type toPropertyType()
        {
            return m_type;
         }
        
        /**
         * Returns the ValueType that exactly matches a given PropertyType;
         * otherwise UNKNOWN.
         * 
         * @param type a Type enumerator
         * @return the ValueType that exactly matches the given Type
         *         PropertyType; otherwise UNKNOWN.
         */
        static public ValueType fromPropertyType(Type type)
        {
            for (ValueType vt: values())
                if (vt.toPropertyType() == type)
                    return vt;
            
            return UNKNOWN;
        }
        
        /**
         * The CqQuery.DisplayField.FieldType that corresponds to this field type.
         * 
         * @return An CqQuery.DisplayField.FieldType enumerator specifying as closely as
         *         possible the Java representation of this type of value
         *         corresponding
         */
        public FieldType toFieldType()
        {
            for(FieldType ft: FieldType.values())
                if (ft.valueType() == this)
                    return ft;
            
            switch (m_type) {
            case RESOURCE:
                return FieldType.REFERENCE;
            case STRING:
                return FieldType.MULTILINE_STRING;
            case OBJECT:
            case RESOURCE_LIST:
                return FieldType.BINARY;
            }
            
            return FieldType.UNKNOWN;
        }
        
        private ValueType(Type type)
        {
            m_type = type;
        }

        private Type m_type;
    }

    /**
     * An enumeration indicating the validity of a field value.
     */
    public enum ValidationStatus
        implements StpExEnumeration
    {
        /** The field value has been validated and found to be valid. */
        VALID,

        /** The field value has been validated and found to be invalid. */
        INVALID,

        /**
         * The field value has not yet been validated and so it is unknown
         * whether it is valid or not.
         */
        NEEDS_VALIDATION;
    }

    /**
     * An enumeration indicating whether or not a field has a value.
     */
    public enum ValueStatus
        implements StpExEnumeration
    {
        /** The field has a value. */
        HAS_A_VALUE,

        /** The field has no value set. */
        HAS_NO_VALUE,

        /**
         * The current state of the field prevents it from returning a value.
         */
        VALUE_NOT_AVAILABLE;
    }

    /**
     * @return the PropertyName of this CqFieldValue property as a
     *         CqRecord.FieldName
     */
    CqRecord.FieldName<T> getFieldName();
    
    /** The record instance of which this field is a part. */
    MetaPropertyName<CqRecord> RECORD = 
        new MetaPropertyName<CqRecord>(METADATA_NAMESPACE, "resource");


    /**
     * Returns the value of the {@link #RECORD RECORD} metadata component as
     * defined by this Property instance.
     *
     * @return  A Record proxy for the record of which this field is a member.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #RECORD RECORD} metadata
     *                         component.
     */
    CqRecord getRecord()
        throws WvcmException;

    /**
     * The field definition that defines all static properties of this field.
     */
    MetaPropertyName<CqFieldDefinition> FIELD_DEFINITION =
        new MetaPropertyName<CqFieldDefinition>(METADATA_NAMESPACE,
                                                "field-definition");

    /**
     * Returns the value of the {@link #FIELD_DEFINITION FIELD_DEFINITION}
     * metadata component as defined by this Property instance.
     *
     * @return  A CqFieldDefinition proxy specifying the static properties of this
     *          field.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #FIELD_DEFINITION FIELD_DEFINITION
     *                         } metadata component.
     */
    CqFieldDefinition getFieldDefinition()
        throws WvcmException;

    /** Identifies the requiredness of the specified field. */
    MetaPropertyName<CqFieldDefinition.Requiredness> REQUIREDNESS =
        new MetaPropertyName<CqFieldDefinition.Requiredness>(METADATA_NAMESPACE,
                                                             "requiredness");

    /**
     * Returns the value of the {@link #REQUIREDNESS REQUIREDNESS} metadata
     * component as defined by this Property instance.
     *
     * @return  A Requiredness enumerator specifying whether this field cannot,
     *          may, or must have a value in the current state.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #REQUIREDNESS REQUIREDNESS}
     *                         metadata component.
     */
    public CqFieldDefinition.Requiredness getRequiredness()
        throws WvcmException;

    /**
     * The current list of possible values for this field. If this list is empty
     * and {@link CqFieldDefinition#HAS_OPEN_CHOICE_LIST} is true, it neither
     * implies that all values are permitted nor that no values are permitted;
     * it just means that the schema designer has not provided any hints about
     * the values permitted in the field. On the other hand, if
     * {@link CqFieldDefinition#HAS_OPEN_CHOICE_LIST} is false then the choice
     * list contains only the values currently permitted for the field.
     * <p>
     * In general, the content of a field's choice list is quite volatile and
     * can depend on the state of the record it is in, the value of other fields
     * of that record or related records, the current user's permissions, and
     * other dynamic aspects of the database environment known only to the
     * schema. So, in general, the choice list must be obtained in the context
     * in which it is to be used at the time it is to be used. In this
     * interface, that very dynamic value is expressed as the value of this
     * meta-property.
     * <p>
     * {@link CqFieldValue#CHOICE_LIST} always returns the choice list for the
     * field whether that list is static or not. If the schema is designed such
     * that the choice list for a field in all records in all contexts is fixed
     * by the schema, then this meta-property will always return the same list
     * for that field. In this static case, the list returned by this
     * meta-property will be the same as the list returned by the
     * {@link CqFieldDefinition#CHOICE_LIST} property.
     * <p>
     * Choice lists can be voluminous and, so, should be read from the server
     * only when needed. Yet, as explained above, they can also be quite
     * volatile and may easily change during the course of a session and not
     * necessarily in direct response to anything the client is doing. To
     * mitigate this problem, a choice-list-id is provided to help the client
     * know when it should request a new copy of the list. Every choice list has
     * a choice-list-id. The value of the choice-list-id changes when and only
     * when the content of the associated choice list changes.
     * <p>
     * If a client does not know that a given field has a static choice list, it
     * must assume that it is not and monitor the field's CHOICE_LIST_ID
     * to know when to load/reload its content.
     */
    MetaPropertyName<java.util.List<String>> CHOICE_LIST =
        new MetaPropertyName<java.util.List<String>>(METADATA_NAMESPACE,
                                                     "choice-list");

    /**
     * Returns the value of the {@link #CHOICE_LIST CHOICE_LIST} metadata
     * component as defined by this Property instance.
     *
     * @return  A List of Strings, each containing a possible value for this
     *          field.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #CHOICE_LIST CHOICE_LIST} metadata
     *                         component.
     */
    java.util.List<String> getChoiceList()
        throws WvcmException;

    /** A String explaining why the value stored in the field is invalid. */
    MetaPropertyName<String> MESSAGE_TEXT =
        new MetaPropertyName<String>(METADATA_NAMESPACE, "message-text");

    /**
     * Returns the value of the {@link #MESSAGE_TEXT MESSAGE_TEXT} metadata
     * component as defined by this Property instance.
     *
     * @return  A String containing any message currently associated with this
     *          field.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #MESSAGE_TEXT MESSAGE_TEXT}
     *                         metadata component.
     */
    String getMessageText()
        throws WvcmException;

    /** Identifies whether the field's value is valid. */
    MetaPropertyName<ValidationStatus> VALIDATION_STATUS =
        new MetaPropertyName<ValidationStatus>(METADATA_NAMESPACE,
                                               "validation-status");

    /**
     * Returns the value of the {@link #VALIDATION_STATUS VALIDATION_STATUS}
     * metadata component as defined by this Property instance.
     *
     * @return  A ValidationStatus enumerator identifying the validation status
     *          of this field.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #VALIDATION_STATUS
     *                         VALIDATION_STATUS} metadata component.
     */
    ValidationStatus getValidationStatus()
        throws WvcmException;

    /** Identifies whether the field currently has a value. */
    MetaPropertyName<ValueStatus> VALUE_STATUS =
        new MetaPropertyName<ValueStatus>(METADATA_NAMESPACE, "value-status");

    /**
     * Returns the value of the {@link #VALUE_STATUS VALUE_STATUS} metadata
     * component as defined by this Property instance.
     *
     * @return  A ValueStatus enumeration indicating whether or not this field
     *          has a value.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #VALUE_STATUS VALUE_STATUS}
     *                         metadata component.
     */
    ValueStatus getValueStatus()
        throws WvcmException;

    /** The type of this field's VALUE meta-property value. */
    public static final MetaPropertyName<ValueType> FIELD_TYPE =
        new MetaPropertyName<ValueType>(METADATA_NAMESPACE, "field-type");

    /**
     * Returns the value of the {@link #FIELD_TYPE FIELD_TYPE} meta-property as
     * defined by this CqFieldValue instance.
     * 
     * @return A VAlueType enumeration representing the type of this property's
     *         VALUE meta-property.
     * 
     * @throws WvcmException if this Property instance does not define a value
     *             for the {@link #FIELD_TYPE FIELD_TYPE} meta-property.
     */
    public ValueType getFieldType()
        throws WvcmException;

    /** True if the field's validity was changed by the current action. */
    MetaPropertyName<Boolean> VALIDITY_CHANGED_THIS_ACTION =
        new MetaPropertyName<Boolean>(METADATA_NAMESPACE,
                                      "validity-changed-this-action");

    /**
     * Returns the value of the {@link #VALIDITY_CHANGED_THIS_ACTION
     * VALIDITY_CHANGED_THIS_ACTION} metadata component as defined by this
     * Property instance.
     *
     * @return  <b>true</b> if the field validity changed during the current
     *          action
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #VALIDITY_CHANGED_THIS_ACTION
     *                         VALIDITY_CHANGED_THIS_ACTION} metadata component.
     */
    boolean getValidityChangedThisAction()
        throws WvcmException;

    /**
     * True if the field's validity was changed by the most recent SetFieldValue
     * call.
     */
    MetaPropertyName<Boolean> VALIDITY_CHANGED_THIS_SET_VALUE =
        new MetaPropertyName<Boolean>(METADATA_NAMESPACE,
                                      "validity-changed-this-set-value");

    /**
     * Returns the value of the {@link #VALIDITY_CHANGED_THIS_SET_VALUE
     * VALIDITY_CHANGED_THIS_SET_VALUE} metadata component as defined by this
     * Property instance.
     *
     * @return  <b>true</b> if the field validity was changed by the most recent
     *          property update.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #VALIDITY_CHANGED_THIS_SET_VALUE
     *                         VALIDITY_CHANGED_THIS_SET_VALUE} metadata
     *                         component.
     */
    boolean getValidityChangedThisSetValue()
        throws WvcmException;

    /**
     * True if the field's validity was changed by the most recent group of
     * SetFieldValue calls.
     */
    MetaPropertyName<Boolean> VALIDITY_CHANGED_THIS_GROUP =
        new MetaPropertyName<Boolean>(METADATA_NAMESPACE,
                                      "validity-changed-this-group");

    /**
     * Returns the value of the {@link #VALIDITY_CHANGED_THIS_GROUP
     * VALIDITY_CHANGED_THIS_GROUP} metadata component as defined by this
     * Property instance.
     * 
     * @return <b>true</b> if this field's validity has changed since the last
     *         setting of {@link CqRecord#FIELDS_UPDATED_THIS_GROUP} in this
     *         field's record.
     * 
     * @throws WvcmException if this Property instance does not define a value
     *             for the {@link #VALIDITY_CHANGED_THIS_GROUP
     *             VALIDITY_CHANGED_THIS_GROUP} metadata component.
     */
    boolean getValidityChangedThisGroup()
        throws WvcmException;

    /** True if this field's value was modified by the current action. */
    MetaPropertyName<Boolean> VALUE_CHANGED_THIS_ACTION =
        new MetaPropertyName<Boolean>(METADATA_NAMESPACE,
                                      "value-changed-this-action");

    /**
     * Returns the value of the {@link #VALUE_CHANGED_THIS_ACTION
     * VALUE_CHANGED_THIS_ACTION} metadata component as defined by this Property
     * instance.
     *
     * @return  <b>true</b> if this field's value has changed during the current
     *          action.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #VALUE_CHANGED_THIS_ACTION
     *                         VALUE_CHANGED_THIS_ACTION} metadata component.
     */
    boolean getValueChangedThisAction()
        throws WvcmException;

    /**
     * True if this field's value was modified by the most recent property
     * update interaction with the server.
     */
    MetaPropertyName<Boolean> VALUE_CHANGED_THIS_SET_VALUE =
        new MetaPropertyName<Boolean>(METADATA_NAMESPACE,
                                      "value-changed-this-set-value");

    /**
     * Returns the value of the {@link #VALUE_CHANGED_THIS_SET_VALUE
     * VALUE_CHANGED_THIS_SET_VALUE} metadata component as defined by this
     * Property instance.
     *
     * @return  <b>true</b> if this fields value changed during the latest
     *          property-update interaction with the server
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #VALUE_CHANGED_THIS_SET_VALUE
     *                         VALUE_CHANGED_THIS_SET_VALUE} metadata component.
     */
    boolean getValueChangedThisSetValue()
        throws WvcmException;

    /**
     * True if the field's value was modified by the most recent group of
     * SetFieldValue calls.
     */
    MetaPropertyName<Boolean> VALUE_CHANGED_THIS_GROUP =
        new MetaPropertyName<Boolean>(METADATA_NAMESPACE,
                                      "value-changed-this-group");

    /**
     * Returns the value of the {@link #VALUE_CHANGED_THIS_GROUP
     * VALUE_CHANGED_THIS_GROUP} metadata component as defined by this Property
     * instance.
     * 
     * @return <b>true</b> if this field's value has changed since the last
     *         setting of {@link CqRecord#FIELDS_UPDATED_THIS_GROUP} in this
     *         field's record.
     * 
     * @throws WvcmException if this Property instance does not define a value
     *             for the {@link #VALUE_CHANGED_THIS_GROUP
     *             VALUE_CHANGED_THIS_GROUP} metadata component.
     */
    boolean getValueChangedThisGroup()
        throws WvcmException;

    /**
     * The field's value as it was retrieved from the database at the start of
     * the action.
     */
    MetaPropertyName<?> ORIGINAL_VALUE =
        MetaPropertyName.build(StpExBase.METADATA_NAMESPACE, "original-value");

    /**
     * Returns the value of the {@link #ORIGINAL_VALUE ORIGINAL_VALUE} metadata
     * component as defined by this Property instance.
     *
     * @return  An Object representing the value of this field just prior to
     *          being made editable in this workspace.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link #ORIGINAL_VALUE ORIGINAL_VALUE}
     *                         metadata component.
     */
    T getOriginalValue()
        throws WvcmException;

    /**
     * The field's value status as it was retrieved from the database at the
     * start of the action.
     */
    MetaPropertyName<ValueStatus> ORIGINAL_VALUE_STATUS =
        new MetaPropertyName<ValueStatus>(StpExBase.METADATA_NAMESPACE, "original-value-status");

    /**
     * Returns the value of the {@link #ORIGINAL_VALUE_STATUS
     * ORIGINAL_VALUE_STATUS} metadata component as defined by this Property
     * instance.
     *
     * @return  A ValueStatus enumerator representing the status of this field's
     *          value just prior to being made editable in this workspace.
     *
     * @throws  WvcmException  if this Property instance does not define a value
     *                         for the {@link
     *                         #ORIGINAL_VALUE_STATUS ORIGINAL_VALUE_STATUS}
     *                         metadata component.
     */
    ValueStatus getOriginalValueStatus()
        throws WvcmException;

    /**
     * Initializes the values of this field value structure.
     *
     * @param  values  A list of value images that are to be the new field value
     *                 images. May be null to indicate a null field value.
     */
    void initialize(java.util.List<String> values);

    /**
     * Initializes the value of this field value structure.
     * 
     * @param  value  The string value that is to be the new field value image.
     *                May be null to indicate a null field value.
     *                
     * @throws WvcmException if the field value is a record or list of records
     * and the specified String is not a valid StpLocation image.
     */
    void initialize(String value) throws WvcmException;

    /**
     * Initializes the value of this field value structure using the value
     * object rather than its image.
     *
     * @param  value  The value of the field. Must be a native representation of
     *                one of the supported field types.
     */
    void initialize(Object value) throws WvcmException;

    /**
     * Returns the number of individual values in an aggregated field value.
     *
     * @return  Returns 0 for a null value, 1 for a scalar value, and the count
     *          of the number of individual values in a aggregated value.
     *          
     * @throws WvcmException if the field value is a record or list of records
     * and the value specified as a String is not a valid StpLocation image.
     */
    long valueCount();

    /**
     * Returns the character image of the i-th element of an aggregated value.
     *
     * @param  i  The (zero-origin) index of the value whose character image is
     *            desired. Must be zero or less than {@link #valueCount()
     *            valueCount}.
     *
     * @return  The character image of the i-th element of the field value. If
     *          the field value is null, {@link #NULL_VALUE_IMAGE
     *          NULL_VALUE_IMAGE} is returned for index 0.
     *
     * @throws  WvcmException  If an i-th value exists in this fields value
     *                         list.
     */
    String characterImage(int i)
        throws WvcmException;

    /**
     * Returns the character image of the (first) field value.
     *
     * @return  Returns {@link #characterImage(int) characterImage(0)}.
     *
     * @throws  WvcmException  If the value is null (valueCount == 0).
     */
    String characterImage()
        throws WvcmException;

    /**
     * The special string value returned by characterImage for a null field
     * value
     */
    String NULL_VALUE_IMAGE = "<CTG:NULL_VALUE/>";
}
